/*
 * B2M_Axes.c
 * Bahn- und Rampenplanung für Beamicon-Modular
 *
 * Created: 13.11.2020 18:37:51
 *  Author: Bene
 */ 


#include "B2M_Axes.h"
#include <math.h>
#include <string.h> // memcpy(), memset()

/* Funktionsweise der Rampenplanung (siehe auch PC-Projekt "RampTest")

Im Puffer werden äquidistante Punkte abgelegt, so als ob mit konstanter
Geschwindigkeit und ohne Rampen gefahren würde. Die Zeiteinheit zwischen zwei
Stützpunkten (aufeinanderfolgenden Puffern) ist 5ms.

Die Hauptschleife der Geschwindigkeitsberechnung läuft mit 1,25ms-Zeitscheiben
(Ticks). Es wird immer zwischen zwei Puffern interpoliert. Bei "voller Fahrt"
viermal pro Puffer (5ms), während der Rampe oder bei Feed-Override entsprechend
langsamer. D.h. die Zeitachse wird gedehnt und ein Puffer wird wesentlich länger
als 5ms verwendet.

Geschwindigkeitswerte werden immer normiert auf eine hypothetische (d.h. nur
dem PC bekannte) Maximalgeschwindigkeit, deren genauer Wert eigentlich hier egal ist.
Die tatsächliche Geschwindigkeit ergibt sich durch den Punktabstand, den der PC
berechnet hat. Der Interpolationszeiger innerhalb des Puffers schreitet mit der
aktuellen Geschwindigkeit voran relativ zur Pufferdistanz, d.h. es werden
vAct / vNom Puffer pro 5ms "verbraucht". vNom ist die Sollgeschwindigkeit und auch
gleichzeitig die normierte Puferdistanz, d.h. die pro 5ms gefahrene Strecke bei
Sollgeschwindigkeit.

Beschleunigung wird in Einheiten der normierten Maximalgeschwindigkeit
pro Tick zum Quadrat angegeben. Z.B. 0.01 bedutet, die Beschleunigung von 0 auf
Sollgeschwindigkeit (1.0) dauert 100 Ticks = 125ms.

Bei der Interpolation zwischen zwei Puffern steht im ersten die Startgeschwindigkeit
und -Position, im zweiten die Endgeschwindigkeit und -Position. Soll auf Null
gebremst werden, wird ein Puffer mit der Endposition und Geschwindigkeit=0
hinzugefügt. Zur einfachen Berechnung der Bremsrampe wird zusätzlich die
verbleibende Reststrecke bis zum Stillstand im Puffereintrag vorberechnet. Achtung,
die Reststrecke gilt vom Anfang der Interpolation (erster Puffer), steht
aber im zweiten Puffer.

Der Feed-Override-Faktor wird mit der Sollgeschwindigkeit multipliziert. Das
Ergebnis darf nicht höher als die Maximalgeschwindigkeit vMax sein. Bei G1-Fahrt
ist dies idR. 150%, bei Eilgang aber nur 100%.

Zahlenformat:

Der P2 hat keine FPU. Für Positionen und digitale Filter sind Integer- bzw.
Fixed-Point-Zahlen aber sowieso besser, weil keine Rundungsfehler auftreten können.
Geschwindigkeiten und Beschleunigungen sind sowieso normiert, haben einen
definierten Wertebereich und passen in 32 Bit. Nur die Reststrecke für die 
Bremsrampe und einige Zwischenergebnisse benötigen 64 Bit. Dies benötigt aber 
auch nicht mehr Speicher als die ursprünglich vogesehenen double Floatingpoint-
Zahlen und wird wesentlich schneller ausgeführt als in Software emulierte
Floatingpoint-Rechnungen.

Geschwindigkeit: 2:30 Bit fixed Point, 2^30 = 1.0, Einheit Puffer/5ms

Beschleunigung:  2:30 Bit fixed Point, Einheit Puffer/(5ms)², 
max. Rampenlänge 2s = 400 Puffer, min. Wert = 0.0025, min. Auflösung 21 Bit

*/

AxisBuffEntry AxisBuffer[AXIS_BUFFER_SIZE];
AxisBuffEntry* buf1; // Interpolation zw. 2 Puffern
AxisBuffEntry* buf2;

fixp32_t currVel; // 1.0
fixp32_t maxVel;  // 1.5
fixp32_t actOvr;  // 1.0
fixp32_t nomAcc;  // 0.05
fixp32_t halfAcc; // 0.025
fixp32_t eStopAcc;// 0.1
fixp32_t vAct; // Ist-Geschwindigkeit
fixp32_t pAct; // interpolierte Position zw. 2 Puffern

int wrIndex;
int rdIndex;
int bufGauge;
int stopFlag; // 0=OK, 1=stop, 2= E-stop
#define stopNorm 1
#define stopFast 2

int32_t nomPos[MAX_NUMAXES]; // ungefilterte Position
int32_t lastPos[MAX_NUMAXES];

/* Ringpuffer Invarianten:
* wrIndex zeigt auf den nächste freien Puffereintrag
* rdIndex zeigt auf den ältesten, gültigen Eintrag (es muss immer min. einer vorhanden sein)
* bufGauge = Anzahl gefüllter Puffereinträge (1 <= bufGauge < AXIS_BUFFER_SIZE)

Damit keine Mutexes nötig sind, ist nur max. ein gleichzeitiger Konsument und ein Produzent
erlaubt. Folgende Reihenfolgen beim zugreifen sind einzuhalten:
* zuerst Puffer schreiben, dann wrIndex und bufGauge erhöhen
* beim Lesen zuerst Puffer lesen dann rdIndex erhöhen, bufGauge verringern

*/

void Init_Axes ()
{
	currVel = 0x40000000; // 1.0
	maxVel  = 0x60000000; // 1.5
	actOvr  = 0x40000000; // 1.0
	nomAcc  = 0x03333333; // 0.05
	halfAcc = 0x01999999; // 0.025
	eStopAcc= 0x06666666; // 0.1
	vAct = 0;
	pAct = 0;
	memset (nomPos, 0, MAX_NUMAXES*4);
	wrIndex = 0;
	rdIndex = 0;
	bufGauge = 0;
	stopFlag = 0;
	SetFilter (1, 1);
	ClearBuffer ();
}

void ClearBuffer ()
{
  currVel= 0;
  PutPosDat (nomPos, GetRefNum ());
  rdIndex= wrIndex - 1;
  buf1= &AxisBuffer[rdIndex];
  buf2= buf1;
  bufGauge= 1;
  pAct= 0;
  vAct= 0;
}

void SetVelocity (fixp32_t v)
{
  currVel= v;
}

void SetAcceleration (fixp32_t aNom, fixp32_t aEstop)
{
  nomAcc= aNom;
  halfAcc= aNom >> 1;
  eStopAcc= aEstop;
}

void SetFeedOverride (fixp32_t ovr)
{
  actOvr= ovr;
  stopFlag= 0;
}

void EmergencyStop ()
{
  stopFlag= stopFast;
  actOvr= 0;
}

int IncIndex (int i)
{
	if (++i >= AXIS_BUFFER_SIZE) return 0;
	return i;
}

fixp32_t FixpMul (fixp32_t a, fixp32_t b)
{

}

void PutPosDat (int32_t* data, int16_t ref)
{
	while (bufGauge >= AXIS_BUFFER_SIZE) {};
    AxisBuffEntry& buf= AxisBuffer[wrIndex]; // neuer Puffereintrag
    memcpy (buf.pos, data, MAX_NUMAXES*4);
    buf.vNom= currVel;
    buf.vMax= maxVel;
	buf.ref = ref;
    uint64_t dist= 0;
	uint64_t v2= halfAcc * halfAcc;
    for (int i=wrIndex; i!=rdIndex; ) // Bremsrampe rückwärts rechnen
    {
	    if (--i < 0) i= AXIS_BUFFER_SIZE - 1;
	    // float v2= 8.0 * dist * nomAcc + (halfAcc * halfAcc); // vDec² vermeidet Wurzel
		fixp32_t v1= AxisBuffer[i].vNom;
	    fixp32_t vMax= v1 + halfAcc;
	    dist+= v1;
	    AxisBuffer[i].dDec= dist;
	    if (v2 > vMax*vMax) break; // vergleiche die Quadrate, geht schneller
		v2+= (v1 * nomAcc) << 3;
    }
    wrIndex= IncIndex (wrIndex);
	bufGauge++;
}

fixp32_t FilterVel (fixp32_t v);
void FilterPos ();

void DoInterpolation ()
{
	if (buf1 == buf2 && bufGauge > 1) // war angehalten
	{
		buf2= &AxisBuffer[IncIndex (rdIndex)]; // wieder loslaufen
	}
    fixp32_t vNom= FixpMul (buf2->vNom, actOvr);
    fixp32_t vMax= stopFlag?0:buf1->vMax;
    if (vMax > 0 && vNom > vMax) vNom= vMax;
    vMax= buf2->vMax;
    if (vMax > 0 && vNom > vMax) vNom= vMax;
    fixp32_t ipol = 0;
    fixp32_t aMax= (stopFlag==stopFast)?eStopAcc:nomAcc; // maximale Verzögerung in einem Tick
    fixp32_t vOld= vAct;
    vAct+= nomAcc; // beschleunigen
    if (vAct > vNom) vAct= vNom; // und begrenzen
    uint64_t brakeDist = buf2->dDec - pAct; // verbleibender Bremsweg
    if (brakeDist < 0) brakeDist= 0;
    fixp32_t vRamp= sqrt (RampVel2 (brakeDist)) - halfAcc;
    if (vAct > vRamp) vAct= vRamp;
    aMax+= vAct - vOld; // Rest der max. möglichen Verzögerung
    if (vAct > vNom || aMax < 0) vAct-= aMax;
    pAct+= FilterVel (vAct) >> 2;
    while (pAct >= buf2->vNom && IncIndex(rdIndex) != wrIndex) // Puffergrenze überschritten
    { // kann mehrmals passieren, aber nur entsprechend dem Override-Faktor (1.5 oder max. 3)
	    pAct-= buf2->vNom; // komplette Distanz abziehen
	    buf1= buf2;
		rdIndex= IncIndex(rdIndex);
		if (--bufGauge > 1) buf2= &AxisBuffer[IncIndex(rdIndex)];
    }
    if (buf2->vNom > 0.0) ipol= pAct / buf2->vNom;
	for (int a=0; a < MAX_NUMAXES; a++)
    { // interpolieren
		nomPos[a]= buf1->pos[a] + round (ipol * (buf2->pos[a] - buf1->pos[a]));
	}
	FilterPos ();
}

/***********************************************************************************/
// Filters

int32_t filtPos[MAX_NUMAXES]; // gefilterte Position
int32_t filtSum[MAX_NUMAXES];
int32_t filtRem[MAX_NUMAXES];
int32_t filtBuff[FILTER_BUFFER_SIZE][MAX_NUMAXES];
fixp32_t filtVel[FILTER_BUFFER_SIZE];
uint64_t velSum;
int filtTimeV;
int filtTimeP;
int filtIdxV;
int filtIdxP;
int idleCntP;

void SetFilter (int vf, int pf)
{
	if (vf > FILTER_BUFFER_SIZE) vf= FILTER_BUFFER_SIZE;
	if (vf < 1) vf= 1;
	if (pf > FILTER_BUFFER_SIZE) pf= FILTER_BUFFER_SIZE;
	if (pf < 1) pf= 1;
	filtTimeV= vf; // Geschwindigkeitsfilter (wirkt auf Rampe, verändert Bahn nicht)
	filtTimeP= pf; // Positionsfilter (auf alle Achsen einzeln, verrundet Bahn)
	filtIdxV= 0;
	idleCntV= 0;
	memset (filtVel, 0, FILTER_BUFFER_SIZE*4);
	velSum= 0;
	filtIdxP= 0;
	idleCntP= 0;
	memset (filtBuff, 0, FILTER_BUFFER_SIZE*4);
	for (int a=0; a<MAX_NUMAXES; a++)
	{
		filtPos[a]= nomPos[a];
		filtSum[a]= 0;
		filtRem[a]= 0;
	}
}

fixp32_t FilterVel (fixp32_t v)
{
	if (++filtIdxV >= filtTimeV) filtIdxV= 0;
	velSum-= filtVel[filtIdxV];
	velSum+= v;
	filtVel[filtIdxV]= v;
	return velSum / filtTimeV;
}

void FilterPos ()
{
	bool moved= false;
	for (int a=0; a<MAX_NUMAXES; a++)
	{
		int32_t diff= nomPos[a] - lastPos[a];
		if (diff) moved= true;
		if (++filtIdxP >= filtTimeP) filtIdxP= 0;
		filtSum[a]+= diff - filtBuff[filtIdxP][a];
		filtBuff[filtIdxP][a]= diff;
		int32_t sum= filtSum[a] + filtRem[a];
		filtPos[a]+= sum / filtTimeP;
		filtRem[a]= sum % filtTimeP;
		lastPos[a]= nomPos[a];
	}
	if (moved) idleCntP= filtTimeP;
	else if (idleCntP) idleCntP--;
}

int16_t GetRefNum ()
{
  return buf1->ref;
}

int32_t GetAxisPos (int a)
{
  return filtPos[a];
}

void GetAllPositions (int32_t* array)
{
  memcpy (array, filtPos, MAX_NUMAXES*4);
}

bool IsMoving ()
{
  return velSum || idleCntP;
}

int GetBufGauge ()
{
  return bufGauge;
}

/***********************************************************************************/
// Test / Debug

void DumpBuffer ()
{
    for (int i=0; i<wrIndex; i++)
    {
        AxisBuffEntry& buf= AxisBuffer[i];
        printf ("buffer[%d] p=%d dn=%f vm=%f dd=%f\n", i, buf.pos, buf.vNom, buf.vMax, buf.dDec);
    }
}

void AddBuffer (int x)
{
	int32_t pos[10];
	pos[0]= x;
	PutPosDat (pos, 0);
}

void Test ()
{
	SetVelocity (0);
    AddBuffer (0);
	SetVelocity (0x40000000);
    for (int i=10; i<=100; i+=10)
        AddBuffer (i*10, 1.0);
	SetVelocity (0x15555555);
    for (int i=103; i<=150; i+=3)
        AddBuffer (i*10);
	SetVelocity (0);
    AddBuffer (1480);
    DumpBuffer ();
    do {
        ExecStep ();
		printf (" Step i=%d pos=%d dp=%d v=%f d=%f vr=%f\n", rdIndex, p, p-lp, vAct, dist, vRamp);
    } while (vAct > 0.001 && rdIndex+1 < wrIndex);
}
